/* thue.c -- Thue equation solver. In all the forthcoming remarks, "paper"
 * designs the paper "Thue Equations of High Degree", by Yu. Bilu and
 * G. Hanrot, * J. Number Theory (1996).
 * The last part of the program (norm.c) was written by K. Belabas.
 */
/* $Id$ */

#include "genpari.h"

static int curne,r,s,t,deg,Prec,flag=0;
static GEN c1,c2,c4,c5,c6,c7,c8,c9,c10,c12,c14,c15,B0,x1,x2,x3,x0;
static GEN halpha,eps3,gdeg,A,MatNE,roo,MatFU,Delta,Lambda,delta,lambda;
static GEN Vect2,SOL,uftot;
GEN normequation(GEN, GEN);

/* Check whether tnf is a valid structure */
static int
checktnf(GEN tnf)
{
  if (typ(tnf)!=t_VEC || (lg(tnf)!=8 && lg(tnf)!=3)) return 0;
  if (typ(tnf[1])!=t_POL) return 0;
  if (lg(tnf) != 8) return 1; /* s=0 */

  deg=lgef(tnf[1])-3; if (deg<=2) err(talker,"Invalid polynomial in thue.");
  s=sturm((GEN)tnf[1]); t=(deg-s)>>1; r=s+t-1;
  checkbnf((GEN)tnf[2]);
  if (typ(tnf[3]) != t_COL || lg(tnf[3]) != deg+1) return 0;
  if (typ(tnf[4]) != t_COL || lg(tnf[4]) != r+1) return 0;
  if (typ(tnf[5]) != t_MAT || lg(tnf[5]) != r+1
      || lg(gmael(tnf,5,1)) != deg+1) return 0;
  if (typ(tnf[6]) != t_MAT || lg(tnf[6])!=r+1
      || lg(gmael(tnf,6,1)) != r+1) return 0;
  if (typ(tnf[7]) != t_VEC || lg(tnf[7]) != 8) return 0;
  return 1;
}

static GEN distoZ(GEN z)
{ 
  GEN p1=gfrac(z);
  return gmin(p1, gsub(gun,p1));
}

/* Check whether a solution has already been found */
static int
new(GEN zz)
{
  int i;
  for (i=1; i<lg(SOL); i++)
    if (gegal(zz,(GEN)SOL[i])) return 0;
  return 1;
}


/* Compensates rounding errors for computation/display of the constants */
static GEN
myround(GEN Cst, GEN upd)
{
  GEN gdix = stoi(10);
  return gmul(Cst,gadd(gun, gmul(upd,gpui(gdix, gneg(gdix), 0))));
}

/* Returns the index of the largest element in a vector */
static int
Vecmax(GEN Vec, int size)
{
  int k, tno = 1;
  GEN tmax = (GEN)Vec[1];
  for (k=2; k<=size; k++)
    if (gcmp((GEN)Vec[k],tmax)==1) { tmax=(GEN)Vec[k]; tno=k; }
  return tno;
}

/* Performs basic computations concerning the equation,
  ie mostly buchinitfu, c1, c2                         */
static void
inithue(GEN poly)
{
  GEN roo2, de, tmp, gpmin;
  int k,j;

  x0=gzero; x1=gzero; deg=lgef(poly)-3; gdeg=stoi(deg);
  uftot=buchinitfu(poly,dbltor(0.2),dbltor(0.2),5); certifybuchall(uftot);
  s=gtolong(gmael3(uftot,7,2,1));
  t=gtolong(gmael3(uftot,7,2,2));
  /* Switch the roots to get the usual order */
  roo=roots(poly, Prec); roo2=cgetg(deg+1,t_COL);
  for (k=1; k<=s; k++) roo2[k]=roo[k];
  for (k=1; k<=t; k++)
  { 
    roo2[k+s]=roo[2*k-1+s];
    roo2[k+s+t]=lconj((GEN)roo2[k+s]);
  }
  roo=roo2;

  r=s+t-1; halpha=gun;
  for (k=1; k<=deg; k++)
    halpha = gmul(halpha,gmax(gun,gabs((GEN)roo[k],Prec)));
  halpha=gdiv(glog(halpha,Prec),gdeg);

  de=deriv(poly,0); c1=gabs(poleval(de,(GEN)roo[1]),Prec);
  for (k=2; k<=s; k++)
  { 
    tmp=gabs(poleval(de,(GEN)roo[k]),Prec);
    if (gcmp(tmp,c1)==-1) c1=tmp;
  }

  c1=gdiv(gpui(gdeux,gsub(gdeg,gun),Prec),c1); c1=myround(c1,gun);
  c2=gabs(gsub((GEN)roo[1],(GEN)roo[2]),Prec);

  for (k=1; k<=deg; k++)
    for (j=k+1; j<=deg; j++)
    {
      tmp=gabs(gsub((GEN)roo[j],(GEN)roo[k]),Prec);
      if (gcmp(c2,tmp)==1) c2=tmp;
    }

  c2=myround(c2,stoi(-1));
  if (t==0) x0=gun;
  else
  {
    gpmin=gabs(poleval(de,(GEN)roo[s+1]),Prec);
    for (k=2; k<=t; k++)
    {
      tmp=gabs(poleval(de,(GEN)roo[s+k]),Prec);
      if (gcmp(tmp,gpmin)==-1) gpmin=tmp;
    }

    /* Compute x0. See paper, Prop. 2.2.1 */
    x0=gpui(gdiv(gpui(gdeux,gsub(gdeg,gun),Prec),
                 gmul(gpmin,
		      gabs((GEN)gimag(roo)[Vecmax(gabs(gimag(roo),Prec),deg)],Prec))),
	    ginv(gdeg),Prec);
  }

  x1=gpui(gdiv(gmul(gadd(gun,gdeg),c1),c2),ginv(gdeg),Prec);
  if (DEBUGLEVEL >=2)
    { fprintferr("c1 = "); outerr(c1); fprintferr("c2 = "); outerr(c2); }
}

/* Computation of M, its inverse A and check of the precision (see paper) */
static void
T_A_Matrices()
{
  GEN mask, eps1, eps2, nia, m1, IntM;
  int i,j;

  m1=glog(gabs(MatFU,Prec),Prec); mask=gpui(gdeux,stoi(r),Prec);
  m1=matextract(m1,gsub(gmul(gdeux,mask),gdeux),gsub(mask,gun));

  A=invmatreel(m1); IntM=gsub(gmul(A,m1),idmat(r));

  eps2=gzero;
  for (i=1; i<=r; i++)
    for (j=1; j<=r; j++)
      if (gcmp(eps2,gabs(gcoeff(IntM,i,j),20)) == -1)
	eps2=gabs(gcoeff(IntM,i,j),20);

  eps1=gpui(gdeux,stoi((Prec-2)*32),Prec); eps2=gadd(eps2,ginv(eps1));

  nia=gzero;
  for (i=1; i<=r; i++)
    for (j=1; j<=r; j++)
      if (gcmp(nia,gabs(gcoeff(A,i,j),20)) == -1)
        nia = gabs(gcoeff(A,i,j),20);

  /* Check for the precision in matrix inversion. See paper, Lemma 2.4.2. */
  if (gcmp(gmul(gadd(gmul(stoi(r),gmul(nia,eps1)),eps2),
           gmul(gdeux,stoi(r))),gun) == -1)
    err(talker,"not enough precision in thue");

  eps3=gmul(gdeux,gmul(gmul(gsqr(stoi(r)),nia),
		       gadd(gmul(stoi(r),gdiv(nia,eps1)),eps2)));
  eps3=gmul(eps3,gpui(gdeg,stoi(s-1),Prec));
  myround(eps3,gun);

  if (DEBUGLEVEL>=2) { fprintferr("epsilon_3 -> "); outerr(eps3); }
}

/* Computation of the constants c5, c7, c10, c12, c15 */
static void
ComputeConstants()
{
  int k;

  GEN Tmp, maxabs, Vect;

  Vect=cgetg(r+1,t_COL); for (k=1; k<=r; k++) Vect[k]=un;

  Delta=gmul(A,Vect);
  c5=(GEN)(gabs(Delta,Prec)[Vecmax(gabs(Delta,Prec),r)]);
  c5=myround(c5,gun); c7=gmul(stoi(r),c5);
  c10=gdiv(gdeg,c7); c15=gdiv(gdeg,c5);
  Tmp=gmul(gabs(A,Prec),Vect); maxabs=(GEN)Tmp[Vecmax(Tmp,r)];
  c12=gmul(gmul(dbltor(2.78),gdiv(c1,c2)),maxabs);

  if (DEBUGLEVEL>=2)
  {
    fprintferr("c5 = "); outerr(c5); fprintferr("c7 = "); outerr(c7);
    fprintferr("c10 = "); outerr(c10); fprintferr("c12 = "); outerr(c12);
    fprintferr("c15 = "); outerr(c15);
  }
}

/* Computation of the constants c6, c8, c9 */
static void
ComputeConstants2()
{
  GEN Vect1, Vect;
  int k;

  Vect=cgetg(r+1,t_COL);
  for (k=1; k<=r; k++) Vect[k]=un;

  Vect1=gdiv(gmul(gabs(A,Prec),Vect),gdeg);
  Vect2=cgetg(r+1,t_COL);
  for (k=1; k<=r; k++)
    Vect2[k]=llog(gabs(gdiv(gsub((GEN)roo[1],(GEN)roo[k+1]),
                            gcoeff(MatNE,k+1,curne)),Prec),Prec);

  Vect1=gadd(Vect1,gabs(gmul(A,Vect2),Prec));
  c6=(GEN)Vect1[Vecmax(Vect1,r)]; c6=myround(c6,gun);
  c8=gadd(dbltor(1.23),gmul(c6,stoi(r)));
  c9=gmul(c4,gexp(gdiv(gmul(gdeg,c8),c7),Prec));
  if (DEBUGLEVEL>=2)
  {
    fprintferr("c6 = "); outerr(c6);
    fprintferr("c8 = "); outerr(c8);
    fprintferr("c9 = "); outerr(c9);
  }
}

/* Computation of the logarithmic heights of the units */
static GEN
Logarithmic_Height(int s)
{
  int i;
  GEN LH=gun,mat;

  mat=gabs(MatFU,Prec);
  for (i=1; i<=deg; i++)
    LH = gmul(LH,gmax(gun,gabs(gcoeff(mat,i,s),Prec)));
  return gmul(gdeux,gdiv(glog(LH,Prec),gdeg));
}

/* Computation of the matrix ((\sigma_i(\eta_j)) used for M_1 and
   the computation of logarithmic heights */
static void
Compute_Fund_Units(GEN uf)
{
  int i,j;
  MatFU=cgetg(r+1,t_MAT);

  for (i=1; i<=r; i++)
    MatFU[i]=lgetg(deg+1,t_COL);
  for (i=1; i<=r; i++)
  {
    if (typ(uf[i])!=t_POL) err(talker, "Incorrect system of units");
    for (j=1; j<=deg; j++)
      coeff(MatFU,j,i)=(long)poleval((GEN)uf[i],(GEN)roo[j]);
  }
}

/* Computation of the conjugates of the solutions to the norm equation */
static void
Conj_Norm_Eq(GEN ne, GEN *Hmu)
{
  int p,q,nesol,x;

  nesol=lg(ne); MatNE=(GEN)cgetg(nesol,t_MAT); (*Hmu)=cgetg(nesol,t_COL);

  for (p=1; p<nesol; p++) { MatNE[p]=lgetg(deg+1,t_COL); (*Hmu)[p]=un; }
  for (q=1; q<nesol; q++)
  {
    x=typ(ne[q]);
    if (x!=t_INT && x!=t_POL)
      err(talker,"Incorrect solutions of norm equation");
    for (p=1; p<=deg; p++)
    {
      coeff(MatNE,p,q)=(long)poleval((GEN)ne[q],(GEN)roo[p]);
      /* Logarithmic height of the solutions of the norm equation */
      (*Hmu)[q]=lmul((GEN)(*Hmu)[q],gmax(gun,gabs(gcoeff(MatNE,p,q),Prec)));
    }
  }
  for (q=1; q<nesol; q++)
    (*Hmu)[q]=ldiv(glog((GEN)(*Hmu)[q],Prec),gdeg);
}

/* Compute Baker's bound c11, and B_0, the bound for the b_i's .*/
static void
Baker(GEN ALH, GEN Hmu)
{
  GEN c11=gun, gbak, hb0=gzero;
  int k;

  gbak = gmul(gmul(gdeg,gsub(gdeg,gun)),gsub(gdeg,gdeux));

  /* For the symbols used for the computation of c11, see paper, Thm 2.3.1 */
  /* Compute h_1....h_r */
  for (k=1; k<=r; k++)
  {
    c11=gmul(c11,gmax((GEN)ALH[k],
		      gmax(ginv(gbak),
			   gdiv(gabs(glog(gdiv(gcoeff(MatFU,2,k),
					       gcoeff(MatFU,3,k)),
					  Prec),Prec),gbak))));
  }

  /* Compute a bound for the h_0 */
  hb0=gadd(gmul(stoi(4),halpha),gadd(gmul(gdeux,(GEN)Hmu[curne]),
                                     gmul(gdeux,glog(gdeux,Prec))));
  hb0=gmax(hb0,gmax(ginv(gbak),
                    gdiv(gabs(glog(gdiv(gmul(gsub((GEN)roo[1],
                                                  (GEN)roo[3]),
                                             gcoeff(MatNE,2,curne)),
                                        gmul(gsub((GEN)roo[1],(GEN)roo[2]),
                                             gcoeff(MatNE,3,curne))),
                                   Prec),Prec),gbak)));
  c11=gmul(c11,hb0);
  /* Multiply c11 by the "constant" factor */
  c11=gmul(c11,gmul(gmul(gmul(stoi(18),gpi),gpui(stoi(32),stoi(4+r),Prec)),
                    gmul(gmul(mpfact(r+3),gpui(gmul(gbak,stoi(r+2)),
                                               stoi(r+3),Prec)),
                         glog(gmul(gdeux,gmul(gbak,stoi(r+2))),Prec))));
  c11=myround(c11,gun);
  /* Compute B0 according to Lemma 2.3.3 */
  B0=gmax(gexp(gun,Prec),
          gmul(gdiv(gmul(gdeux,c11),c10),
	       glog(gdiv(gmul(c11,gpui(c9,ginv(c11),Prec)),c10),Prec)));

  if (DEBUGLEVEL>=2)
    { fprintferr("Baker -> "); outerr(c11); fprintferr("B0 -> "); outerr(B0); }
}

/* Compute delta and lambda */
static void
Create_CF_Values(int *i1, int *i2, GEN *errdelta)
{
  GEN eps5;
  int j1,j2;

  j1=Vecmax(gabs(Delta,Prec),r);
  if (j1!=1) j2=1; else j2=2;

  delta=gdiv((GEN)Delta[j2],(GEN)Delta[j1]);
  eps5=gmul(eps3,stoi(r));
  *errdelta=gmul(gadd(gun,delta),
		 gdiv(eps5,gsub(gabs((GEN)Delta[j1],Prec),eps5)));

  Lambda=gmul(A,Vect2);
  lambda=gdiv(gsub(gmul((GEN)Delta[j2],(GEN)Lambda[j1]),
                   gmul((GEN)Delta[j1],(GEN)Lambda[j2])),
	      (GEN)Delta[j1]);
  c14=gmul(c12,gexp(gmul(gdeg,gdiv(c6,c5)),Prec));

  if (DEBUGLEVEL>=2)
  {
    fprintferr("c12 = "); outerr(c12); fprintferr("Imprecision on delta -> ");
    outerr(*errdelta);
    fprintferr("c14 = "); outerr(c14); fprintferr("c15 = "); outerr(c15);
  }
  *i1=j1; *i2=j2;
}

/* Try to reduce the bound through continued fractions; see paper. */
static int
CF_First_Pass(GEN kappa, GEN errdelta)
{
  GEN q, ql,gdix;

  gdix=stoi(10);
  if (gcmp(gmul(gdiv(gun,gdix),gsqr(gmul(B0,kappa))),ginv(errdelta))==1)
    err(talker,"not enough precision in thue");

  q=denom(bestappr(delta, gmul(kappa, B0))); ql=gmul(q,lambda);
  if (gcmp(gabs(gsub(ql, ground(ql)), Prec), gdiv(gdeux, kappa))==1)
  {
    B0=gdiv(gadd(glog(B0, Prec),
                 glog(gmul(gdiv(gmul(gdix,c14),stoi(9)), gsqr(kappa)), Prec)),
	    c15);

    if (DEBUGLEVEL>=2)
    {
      fprintferr("CF_First_Pass successful !!\n");
      fprintferr("B0 -> "); outerr(B0);
    }
    return 1;
  }
  if (DEBUGLEVEL>=2)
    fprintferr("CF_First_Pass failed. Trying again with larger kappa\n");

  return 0;
}

/* Check whether a "potential" solution is truly a solution. */
static void
Check_Solutions(GEN xmay1, GEN xmay2, GEN poly, GEN rhs)
{
  GEN xx1, xx2, yy1, yy2, zz, u;

  yy1=ground(greal(gdiv(gsub(xmay2,xmay1), gsub((GEN)roo[1],(GEN)roo[2]))));
  yy2=gneg(yy1);

  xx1=greal(gadd(xmay1,gmul((GEN)roo[1],yy1)));
  xx2=gneg(xx1);

  if (gcmp(distoZ(xx1),dbltor(0.0001))==-1)
  {
    xx1=ground(xx1);
    if (!gcmp0(yy1))
    {
      if (gegal(gmul(poleval(poly,gdiv(xx1,yy1)),
		     gpui(yy1,gdeg,Prec)),(GEN)rhs))
      {
	zz=cgetg(2,t_VEC); u=cgetg(3,t_VEC);
	u[1]=(long)xx1; u[2]=(long)yy1; zz[1]=(long)u;
	if (new(u)) SOL=concat(SOL,zz);
      }
    }
    else if (gegal(gpui(xx1,gdeg,Prec),(GEN)rhs))
    {
      zz=cgetg(2,t_VEC); u=cgetg(3,t_VEC);
      u[1]=(long)xx1; u[2]=(long)gzero; zz[1]=(long)u;
      if (new(u)) SOL=concat(SOL,zz);
    }

    xx2=ground(xx2);
    if (!gcmp0(yy2))
    {
      if (gegal(gmul(poleval(poly,gdiv(xx2,yy2)),
		     gpui(yy2,gdeg,Prec)),(GEN)rhs))
      {
	zz=cgetg(2,t_VEC); u=cgetg(3,t_VEC);
	u[1]=(long)xx2; u[2]=(long)yy2; zz[1]=(long)u;
	if (new(u)) SOL=concat(SOL,zz);
      }
    }
    else if (gegal(gpui(xx2,gdeg,Prec),(GEN)rhs))
    {
      zz=cgetg(2,t_VEC); u=cgetg(3,t_VEC);
      u[1]=(long)xx2; u[2]=(long)gzero; zz[1]=(long)u;
      if (new(u)) SOL=concat(SOL,zz);
    }
  }
}

/* Check for solutions under a small bound (see paper) */
static void
Check_Small(int bound, GEN poly, GEN rhs)
{
  GEN interm, possy, xx, zz, u;
  int nbpossy, x, i, j;

  for (x=-bound; x<=bound; x++)
  {
    xx=stoi(x); interm=gsub((GEN)rhs,gpui(xx,gdeg,Prec));
    if (gcmp(gabs(interm,Prec), stoi(100000000))==-1 && !gegal(interm,gzero))
    {
      possy=divisors(gabs(interm,Prec)); nbpossy=lg(possy)-1;
      for (j=1; j<=nbpossy; j++)
      {
	if (gegal(gmul(poleval(poly,gdiv(xx,(GEN)possy[j])),
		       gpui((GEN)possy[j],gdeg,Prec)),(GEN)rhs))
	{
	  zz=cgetg(2,t_VEC); u=cgetg(3,t_VEC);
	  u[1]=(long)xx; u[2]=possy[j]; zz[1]=(long)u;
	  if (new(u)) SOL=concat(SOL,zz);
	}
	if (gegal(gmul(poleval(poly,gdiv(xx,gneg((GEN)possy[j]))),
	               gpui(gneg((GEN)possy[j]),gdeg,Prec)), (GEN)rhs))
	{
	  zz=cgetg(2,t_VEC); u=cgetg(3,t_VEC);
	  u[1]=(long)xx; u[2]=lneg((GEN)possy[j]); zz[1]=(long)u;
	  if (new(u)) SOL=concat(SOL,zz);
	}
      }
    }
    else
    {
      GEN newpol = polrecip(gsubst(poly,0,gmul(polx[0],xx)));

      newpol=gsub(newpol,rhs); possy=roots(newpol,3);
      for (i=1; i<=deg; i++)
      {
	if (gcmp(distoZ(greal((GEN)possy[i])), dbltor(0.0001))==-1)
	{
	  if (gegal(poleval(newpol,ground(greal((GEN)possy[i]))),gzero))
	  {
	    zz=cgetg(2,t_VEC); u=cgetg(3,t_VEC);
	    u[1]=(long)xx; u[2]=(long)ground(greal((GEN)possy[i]));
	    zz[1]=(long)u;
	    if (new(u)) SOL=concat(SOL,zz);
	  }
	}
      }
    }
  }
}

/* Computes [x]! */
static double
fact(double x)
{ 
  double ft=1.0;
  x=(int)x; while (x>1) { ft*=x; x--; }
  return ft ;
}

/* From a polynomial and optionally a system of fundamental units for the
 * field defined by poly, computes all the relevant constants needed to solve
 * the equation P(x,y)=a whenever the solutions of N_{ K/Q }(x)=a.  Returns a
 * "tnf" structure containing
 *  1) the polynomial
 *  2) the bnf (used to solve the norm equation)
 *  3) roots, with presumably enough precision
 *  4) The logarithmic heights of units
 *  5) The matrix of conjugates of units
 *  6) its inverse
 *  7) a few technical constants
 */
GEN
thueinit(GEN poly)
{
  GEN thueres,ALH,csts,c0;
  int ltop,lbot,k,st;
  double d,dr;

  ltop=avma;
  if (checktnf(poly)) { uftot=(GEN)poly[2]; poly=(GEN)poly[1]; }
  else
    if (typ(poly)!=t_POL)
      err(talker, "Not a polynomial in thueinit");

  if (!gisirreducible(poly))
    err(talker,"Reducible polynomial in thueinit");
  st=sturm(poly);
  if (st)
  {
    if (!flag)
    {
      dr=(double)((st+lgef(poly)-5)>>1);
      d=(double)lgef(poly)-3; d=d*(d-1)*(d-2);
     /* Try to guess the precision by approximating Baker's bound.
      * Note that the guess is most of the time pretty generous,
      * ie 10 to 30 decimal digits above what is *really* necessary.
      * Note that the limiting step is the reduction. See paper.
      */
      Prec=3 + (long)((5.83 + (dr+4)*5 + log(fact(dr+3)) + (dr+3)*log(dr+2) +
             (dr+3)*log(d) + log(log(2*d*(dr+2))) + (dr+1)) / 11.09);
    }
    if (!checktnf(poly)) inithue(poly);

    thueres=cgetg(8,t_VEC);
    thueres[1]=(long)poly;
    thueres[2]=(long)uftot;
    thueres[3]=(long)roo;
    Compute_Fund_Units(gmael(uftot,8,5));
    ALH=cgetg(r+1,t_COL);
    for (k=1; k<=r; k++) ALH[k]=(long)Logarithmic_Height(k);
    thueres[4]=(long)ALH;
    thueres[5]=(long)MatFU;
    T_A_Matrices();
    thueres[6]=(long)A;

    csts=cgetg(8,t_VEC);
    csts[1]=(long)c1; csts[2]=(long)c2; csts[3]=(long)halpha;
    csts[4]=(long)x0; csts[5]=(long)x1; csts[6]=(long)eps3;
    csts[7]=(long)stoi(Prec);

    thueres[7]=(long)csts; lbot=avma;
    return gerepile(ltop,lbot,gcopy(thueres));
  }

  thueres=cgetg(3,t_VEC); c0=gun; Prec=4;
  roo=roots(poly,Prec);
  for (k=1; k<lg(roo); k++)
    c0=gmul(c0, gimag((GEN)roo[k]));
  c0=ginv(c0);
  thueres[1]=(long)poly; thueres[2]=(long)c0;
  lbot=avma; return gerepile(ltop,lbot,gcopy(thueres));
}

/* Given a tnf structure as returned by thueinit, and a right-hand-side and
 * optionally the solutions to the norm equation, returns the solutions to
 * the Thue equation F(x,y)=a
 */
GEN
thue(GEN thueres, GEN rhs, GEN ne)
{
  GEN Kstart,Old_B0,rootmp,T,ALH,errdelta,Hmu,c0,poly,csts;
  GEN gm1, gm2,c3, bet1,bet0,xmay1,xmay2,b,b1;
  int Nb_It=0,k,upb,bi1,ltop,lbot,i1,i2,numroot,i,j;

  if (!checktnf(thueres)) err(talker,"Not a tnf in thue");

  ltop=avma; SOL=cgetg(1,t_VEC);
  if (lg(thueres)==8)
  {
    poly=(GEN)thueres[1]; deg=lgef(poly)-3; gdeg=stoi(deg);
    uftot=(GEN)thueres[2];
    roo=gcopy((GEN)thueres[3]);
    ALH=(GEN)thueres[4];
    MatFU=gcopy((GEN)thueres[5]);
    A=gcopy((GEN)thueres[6]);
    csts=(GEN)thueres[7];

    if (gcmp0(ne))
      ne = normequation(uftot, rhs);

    c1=gmul(gabs(rhs,Prec), (GEN)csts[1]); c2=(GEN)csts[2];
    halpha=(GEN)csts[3];
    x0=gmul((GEN)csts[4],gpui(gabs(rhs,Prec), ginv(gdeg), Prec));
    x1=gmax(x0,(GEN)csts[5]); eps3=(GEN)csts[6];
    Prec=gtolong((GEN)csts[7]); x2=gzero; x3=gzero; gm1=stoi(-1);
    gm2=stoi(-2); b=cgetg(r+1,t_COL);

    /* Case r=1 is different, since one cannot choose i1!=i2 when
       reducing Baker's bound, for instance.
       However one can deal with it by using (15) properly */
    if (r==1)
    {
      bet1=gdiv(gcoeff(MatFU,2,1),gcoeff(MatFU,3,1));
      for (curne=1; curne<lg(ne); curne++)
      {
	bet0=gmul(gdiv(gsub((GEN)roo[1],(GEN)roo[3]),
	               gsub((GEN)roo[1],(GEN)roo[2])),
		  gdiv(poleval((GEN)ne[curne],(GEN)roo[2]),
		       poleval((GEN)ne[curne],(GEN)roo[3])));
	b1=ground(gdiv(glog(gabs(bet0,Prec),Prec),glog(gabs(bet1,Prec),Prec)));
	xmay1=gpui(gcoeff(MatFU,1,1),(GEN)b1,Prec);
	xmay2=gpui(gcoeff(MatFU,2,1),(GEN)b1,Prec);
	xmay1=gmul(xmay1,poleval((GEN)ne[curne],(GEN)roo[1]));
	xmay2=gmul(xmay2,poleval((GEN)ne[curne],(GEN)roo[2]));
	Check_Solutions(xmay1,xmay2,poly,rhs);
      }
      c3=gmul(c1,gmul(gdiv(gadd(gdeg,gun),gdeg),
		      gabs(gdiv(gsub((GEN)roo[3],(GEN)roo[2]),
				gmul(gsub((GEN)roo[3],(GEN)roo[1]),
				     gsub((GEN)roo[1],(GEN)roo[2]))),Prec)));
      c3=myround(c3,gun);
      Check_Small(gtolong(ground(gpui(gmul(gdeux,c3),ginv(gdeg),Prec))),poly,rhs);
      lbot=avma; return gerepile(ltop,lbot,gcopy(SOL));
    }

    T=cgetg(s+t,t_MAT);
    for (k=1; k<=s+t-1; k++) T[k]=lgetg(s+t,t_COL);

    for (k=1; k<=s-1; k++) coeff(T,1,k) = (int) gm1;
    for (k=s; k<=s+t-1; k++) coeff(T,1,k) = (int) gm2;
    for (i=2; i<=s+t-1; i++)
      for (j=1; j<=s+t-1; j++)
	coeff(T,i,j) = (i==j) ? un : zero;

    for (numroot=1; numroot<=s; numroot++)
    {
      if (numroot > 2)
      {
	for (k=1; k<=s-1; k++)
	{
	  coeff(T,numroot-1,k)=(int)gm1;
	  coeff(T,numroot-2,k)=zero;
	}
	for (   ; k<=s+t-1; k++)
	{
	  coeff(T,numroot-1,k)=(int)gm2;
	  coeff(T,numroot-2,k)=zero;
	}
	coeff(T, numroot-2, numroot-2)=un;
      }
      if (numroot>=2)
      {
	/* Compute the new A matrix */
	A=gmul(A,T);

	/* Swap root no 1 (which is in fact root no numroot-1) and
	   root numroot. Do the same for MatFU and MatNE. */
	rootmp=(GEN)roo[numroot]; roo[numroot]=roo[1];
	roo[1]=(long)rootmp;
	for (k=1; k<=r; k++)
	{
	  rootmp=gcoeff(MatFU,numroot,k);
	  coeff(MatFU,numroot,k)=coeff(MatFU,1,k);
	  coeff(MatFU,1,k)=(long)rootmp;
	}

	for (k=1; k<lg(ne); k++)
	{
	  rootmp=gcoeff(MatNE,numroot,k);
	  coeff(MatNE,numroot,k)=coeff(MatNE,1,k);
	  coeff(MatNE,1,k)=(long)rootmp;
	}
      }
      ComputeConstants();

      c3=gmul(c1,gmul(gdiv(gadd(gdeg,gun),gdeg),
		      gabs(gdiv(gsub((GEN)roo[3],(GEN)roo[2]),
				gmul(gsub((GEN)roo[3],(GEN)roo[1]),
				     gsub((GEN)roo[1],(GEN)roo[2]))),Prec)));
      c3=myround(c3,gun); c4=gmul(dbltor(1.39),c3);
      x2=gmax(x1,gpui(gmul(c3,gdeux),ginv(gdeg),Prec));

      Conj_Norm_Eq(ne,&Hmu);
      for (curne=1; curne<lg(ne); curne++)
      {
	ComputeConstants2();
	Baker(ALH,Hmu);
	Create_CF_Values(&i1,&i2,&errdelta);
	if (DEBUGLEVEL>=2) fprintferr("Entering CF\n");
	Old_B0=gadd(B0,gun);

    /* Try to reduce B0 while
     * 1) it is larger than 20
     * 2) there was less than 10 reductions
     * 3) the previous reduction improved the bound
     */
	while (gcmp(B0,stoi(20)) == 1 && Nb_It < 10 && gcmp(Old_B0,B0) == 1)
	{
	  Old_B0=B0; Nb_It++; Kstart=stoi(10);
	  while (!CF_First_Pass(Kstart,errdelta))
	    Kstart=mulis(Kstart,10);
	}
	Nb_It=0; B0=gmin(Old_B0,B0); upb=gtolong(gceil(B0));
	x3=gmax(x2,gpui(gmul(gmul(gdeux,c1),c15),ginv(gdeg),Prec));

	if (DEBUGLEVEL >=2) { fprintferr("x3 -> "); outerr(x3); }

      /* For each possible value of b_i1, compute the b_i's
       * and 2 conjugates of x-\alpha y. Then check.
       */
	for (bi1=-upb; bi1<=upb; bi1++)
	{
	  xmay1=gun; xmay2=gun;
	  for (i=1; i<=r; i++)
	  {
	    b[i]=lround(gdiv(gsub(gmul((GEN)Delta[i],stoi(bi1)),
	                          gsub(gmul((GEN)Delta[i],(GEN)Lambda[i1]),
				       gmul((GEN)Delta[i1],(GEN)Lambda[i]))),
			     (GEN)Delta[i1]));
	    xmay1=gmul(xmay1,gpui(gcoeff(MatFU,1,i),(GEN)b[i],Prec));
	    xmay2=gmul(xmay2,gpui(gcoeff(MatFU,2,i),(GEN)b[i],Prec));
	  }
	  xmay1=gmul(xmay1,gcoeff(MatNE,1,curne));
	  xmay2=gmul(xmay2,gcoeff(MatNE,2,curne));
	  Check_Solutions(xmay1,xmay2,poly,rhs);
	}
      }
      if (DEBUGLEVEL>=2)
      {
	fprintferr("Now checking for small solutions\n");
	Check_Small((int)(gtodouble(x3)),poly,rhs);
      }
    }
    lbot=avma; flag=0; return gerepile(ltop,lbot,gcopy(SOL));
  }

  /* Case s=0. All the solutions are "small". */
  Prec=4; poly=(GEN)thueres[1]; deg=lgef(poly)-3;
  gdeg=stoi(deg); c0=(GEN)thueres[2];
  Check_Small(gtolong(ground(gpui(gmul((GEN)poly[2],gdiv(gabs(rhs,Prec),c0)),
                                  ginv(gdeg),Prec) )), poly, rhs);
  flag=0; lbot=avma; return gerepile(ltop,lbot,gcopy(SOL));
}

static GEN *Relations; /* primes above a, expressed on generators of Cl(K) */
static GEN *Partial;   /* list of vvectors, Partial[i] = Relations[1..i] * u[1..i] */
static GEN *gen_ord;   /* orders of generators of Cl(K) given in bnf */

static long *f;        /* f[i] = f(Primes[i]/p), inertial degree */
static long *n;        /* a = prod p^{ n_p }. n[i]=n_p if Primes[i] divides p */
static long *inext;    /* index of first P above next p, 0 if p is last */
static long *S;        /* S[i] = n[i] - sum_{ 1<=k<=i } f[k].u[k] */
static long *u;        /* We want principal ideals I = prod Primes[i]^u[i] */
static long **normsol; /* lists of copies of the u[] which are solutions */

static long Nprimes; /* length(Relations) = #{max ideal above divisors of a} */
static long sindex, smax; /* current index in normsol; max. index */

static void loop(long i);

GEN
normequation(GEN bnf, GEN a)
{
  GEN nf=checknf(bnf), pol=(GEN)nf[1];
  GEN fact, primes, res, *Primes, *Fact;
  long nprimes, lbot, ltop=avma;
  long i,j,ok;
  long Ngen; /* height(Relations) = #{classgroup generators} */

  if (typ(a)!=t_INT)
    err(talker,"expected an integer in normequation");

  fact=factor(absi(a)); primes=(GEN)fact[1]; nprimes=lg(primes)-1;
  if (!nprimes) { avma=ltop; res=cgetg(2,t_VEC); res[1]=un; return res; }

  gen_ord = (GEN *) mael3(bnf,8,1,2); Ngen = lg(gen_ord)-1;

  Fact = (GEN*) xmalloc((1+nprimes)*sizeof(GEN));
  Nprimes=0;
  for (i=1; i<=nprimes; i++)
  {
    GEN dec = primedec(bnf,(GEN)primes[i]);

    /* if only one prime above p, we must have f_p | n_p */
    if (lg(dec) == 2)
      if (! divise(gmael(fact,2,i), gmael(dec,1,4)))
      {
	free(Fact); avma=ltop;
	res=cgetg(2,t_VEC); res[1]=un; return res;
      }
    Fact[i] = dec;
    Nprimes += lg(dec)-1;
  }

  f = (long*) xmalloc((1+Nprimes)*sizeof(long));
  u = (long*) xmalloc((1+Nprimes)*sizeof(long));
  n = (long*) xmalloc((1+Nprimes)*sizeof(long));
  S = (long*) xmalloc((1+Nprimes)*sizeof(long));
  inext = (long*) xmalloc((1+Nprimes)*sizeof(long));

  if (Ngen)
  {
    Partial = (GEN*) xmalloc((1+Nprimes)*sizeof(GEN));
    Relations = (GEN*) xmalloc((1+Nprimes)*sizeof(GEN));
  }
  else /* trivial class group, no relations to check */
    Relations = Partial = NULL;

  j=0;
  for (i=1; i<=nprimes; i++)
  {
    GEN dec = Fact[i];
    long k, lim = lg(dec);
    long v = (i==nprimes)? 0 : j + lim, vn=itos(gmael(fact,2,i));

    for (k=1; k < lim; k++)
    {
      j++; Primes[j] = (GEN)dec[k];
      f[j] = itos(gmael(dec,k,4));
      n[j] = vn;
      inext[j] = v;
      if (Partial)
	Relations[j]=isprincipal(bnf,(GEN)dec[k]);
    }
  }
  free(Fact);

  if (Partial)
    for (i=0; i<=Nprimes; i++) /* Partial[0] needs to be initialized */
    {
      Partial[i]=cgetg(Ngen+1,t_COL);
      for (j=1; j<=Ngen; j++)
      {
	Partial[i][j]=lgeti(4);
	gaffect(gzero,(GEN) Partial[i][j]);
      }
    }
  sindex=0; smax=511;
  normsol = (long **) xmalloc((smax+1)*sizeof(long *));
  S[0]=n[1]; inext[0]=1;loop(0);

  res=cgetg(1,t_VEC);
  for (i=1; i<=sindex; i++)
  {
    GEN x, id=gun;
    long *uu=normsol[i];

    for (j=1; j<=Nprimes; j++) /* compute prod Primes[i]^u[i] */
      if (uu[j])
      {
	GEN id2 = Primes[j];

	if (uu[j] != 1) id2 = idealpow(nf,id2, stoi(uu[j]));
	id = idealmul(nf,id,id2);
      }
    x = (GEN) isprincipalgen(bnf,id)[2];
    x = gmul((GEN)nf[7],x); ok=1;
    if ( signe(a) != signe(gnorm(gmodulcp(x,(GEN)nf[1]))))
    {
      GEN fu=gmael(bnf,8,5);
      ok=0;
      for (j=1; j<lg(fu); j++)
      {
        const GEN unit = (GEN)fu[j];
        if (signe(gnorm(gmodulcp(unit,pol))) == -1)
        {
          x=gmul(unit,x); ok=1; break;
        }
      }
    }
    if (ok) res=concat(res,gmod(x,pol));
    else if (DEBUGLEVEL > 2)
    {
      output(res);
      fprintferr(" eliminated because of sign");
      flusherr();
    }
    free(uu);
  }
  free(f); free(u); free(S); free(n);
  free(Primes); free(inext);
  free(gen_ord); free(normsol);
  if (Partial) { free(Relations); free(Partial); }

  if (DEBUGLEVEL > 2)
    for (i=1; i<lg(res); i++)
      output(gnorm(gmodulcp((GEN)res[i],pol)));
  
  lbot=avma; return gerepile(ltop,lbot,gcopy(res));
}

/* Recursive loop. Suppose u[1..i] has been filled
 * Find possible solutions u such that, Norm(prod Prime[i]^u[i]) = a, taking 
 * into account:
 *  1) the relations in the class group if need be.
 *  2) the factorization of a.
 */
static void
loop(long i)
{
  int do_test=0, k;

  if (S[i] == 0) /* sum u[i].f[i]=n[i] --> do another prime */
  {
    if (inext[i]) /* there are some primes left */
    {
      if (Partial) gaffect(Partial[i], Partial[inext[i]-1]);
      for (k=i+1; k < inext[i]; k++) u[k]=0;
      i=inext[i]-1;
    }
    else do_test=1;
  }
  else if (i == inext[i]-2 || i == Nprimes-1)
  {
    /* only one Prime left above this given prime; change prime, fix u[i+1] */
    if (S[i] % f[i+1]) return;
    i++; u[i]=S[i-1]/f[i];
    if (Partial)
      for (k=1; k<=Nprimes; k++)
        addiiz(
          (GEN) Partial[i-1][k],
                mulis((GEN) Relations[i][k], u[i]),
          (GEN) Partial[i][k]
        );
    do_test = (inext[i]==0);
  }

  if (do_test)
  {
    long *sol;

    if (Partial)
      for (k=1; k<=Nprimes; k++)
        if ( signe(modii( (GEN)Partial[i][k], gen_ord[k] )) ) return;
    if (sindex == smax)
    {
      long new_smax = smax << 1;
      normsol = (long **) xrealloc(normsol, (smax+1)*sizeof(long *),
                                  (new_smax+1)*sizeof(long *));
      smax=new_smax;
    }
    sol = (long *) xmalloc((Nprimes+1)*sizeof(long));
    normsol[++sindex] = sol;

    for (k=1; k<=i; k++) sol[k]=u[k];
    for (   ; k<=Nprimes; k++) sol[k]=0;
    if (DEBUGLEVEL > 2)
    {
      fprintferr("[");
      for (k=1; k<=Nprimes; k++)
      {
	fprintferr("%ld",sol[k]);
	if (k<Nprimes) fprintferr(",");
      }
      fprintferr("]\n"); flusherr();
    }
    return;
  }

  if (Partial && i) gaffect(Partial[i-1], Partial[i]);
  i++;
  if (i == inext[i-1])
  { /* change prime */
    if (inext[i] == i+1 || i == Nprimes) /* only one Prime above p */
    {
      S[i]=0;
      u[i] = (n[i] / f[i]); /* we already know this is exact */
    }
    else
      {u[i] = 0; S[i] = n[i];}
  } 
  else
    S[i] = S[i-1]; /* same prime, different Prime */
  for(;;)
  {
    loop(i); S[i]-=f[i]; if (S[i]<0) break;
    if (Partial)
      gaddz(Partial[i], Relations[i], Partial[i]);
    u[i]++;
  }
}
